import os

import nox
from nox import Session, param, parametrize, session

nox.options.error_on_external_run = True
nox.options.default_venv_backend = "uv"
nox.options.sessions = [
    "lint",
    "type_check",
    "typos",
    "test",
]


@session(python=["3.11", "3.12"])
def test(s: Session) -> None:
    s.install(".[data,llm,statistics,viz]", "pytest", "pytest-asyncio", "pytest-cov", "pytest-randomly")
    s.run_install(
        "uv",
        "sync",
        "--locked",
        "--all-extras",
        env={"UV_PROJECT_ENVIRONMENT": s.virtualenv.location},
    )

    # Skip tests in directories specified by the SKIP_TESTS_NAII environment variable.
    skip_tests = os.getenv("SKIP_TESTS_NAAI", "")
    skip_tests += " tests/llm/chat_completion/ tests/llm/embedding/"
    skip_args = [f"--ignore={dir}" for dir in skip_tests.split()] if skip_tests else []

    s.run(
        "python",
        "-m",
        "pytest",
        "--cov=not_again_ai",
        "--cov-report=html",
        "--cov-report=term",
        "tests",
        *skip_args,
        "-W ignore::DeprecationWarning",
        *s.posargs,
    )


# For some sessions, set venv_backend="none" to simply execute scripts within the existing
# uv-generated virtual environment, rather than have nox create a new one for each session.
@session(venv_backend="none")
@parametrize(
    "command",
    [
        param(
            [
                "ruff",
                "check",
                ".",
                "--select",
                "I",
                # Also remove unused imports.
                "--select",
                "F401",
                "--extend-fixable",
                "F401",
                "--fix",
            ],
            id="sort_imports",
        ),
        param(["ruff", "format", "."], id="format"),
    ],
)
def fmt(s: Session, command: list[str]) -> None:
    s.run(*command)


@session(venv_backend="none")
@parametrize(
    "command",
    [
        param(["ruff", "check", "."], id="lint_check"),
        param(["ruff", "format", "--check", "."], id="format_check"),
    ],
)
def lint(s: Session, command: list[str]) -> None:
    s.run(*command)


@session(venv_backend="none")
def lint_fix(s: Session) -> None:
    s.run("ruff", "check", ".", "--extend-fixable", "F401", "--fix")


@session(venv_backend="none")
def type_check(s: Session) -> None:
    s.run("mypy", "src", "tests", "noxfile.py")


@session(venv_backend="none")
def typos(s: Session) -> None:
    s.run("typos", "-c", ".github/_typos.toml")



import pytest

from not_again_ai.llm.chat_completion.types import MessageT, SystemMessage, UserMessage
from not_again_ai.llm.prompting import Tokenizer


@pytest.fixture(
    params=[
        {"model": "gpt-4o-mini-2024-07-18", "provider": "openai"},
        {"model": "o1-mini", "provider": "azure_openai"},
    ]
)
def tokenizer(request: pytest.FixtureRequest) -> Tokenizer:
    return Tokenizer(**request.param)


@pytest.fixture(
    params=[
        {"model": "gpt-4o-mini-2024-07-18", "provider": "openai"},
        {"model": "o1-mini", "provider": "azure_openai"},
        {"model": "gpt-4o-mini-2024-07-18", "provider": "mistral"},
        {"model": "unknown-model", "provider": "openai"},
    ]
)
def tokenizer_with_unsupported(request: pytest.FixtureRequest) -> Tokenizer:
    return Tokenizer(**request.param)


# This test only runs with OpenAI and Azure
def test_truncate_str(tokenizer_with_unsupported: Tokenizer) -> None:
    text = "This is a test sentence for the function."
    max_len = 3
    result = tokenizer_with_unsupported.truncate_str(text, max_len)
    print(result)


# This test runs with all three providers
def test_truncate_str_no_truncation(tokenizer: Tokenizer) -> None:
    text = "Short text"
    max_len = 20
    result = tokenizer.truncate_str(text, max_len)
    print(result)


def test_num_tokens_in_str(tokenizer_with_unsupported: Tokenizer) -> None:
    text = "This is a test sentence for the function."
    result = tokenizer_with_unsupported.num_tokens_in_str(text)
    print(result)


def test_num_tokens_allowed_special() -> None:
    tokenizer = Tokenizer(
        model="gpt-4o-mini-2024-07-18", provider="openai", allowed_special=set(), disallowed_special=()
    )
    text = "<|endoftext|>"
    result = tokenizer.num_tokens_in_str(text)
    print(result)


def test_num_tokens_multiple_messages(tokenizer_with_unsupported: Tokenizer) -> None:
    messages: list[MessageT] = [SystemMessage(content="System message."), UserMessage(content="User message.")]
    result = tokenizer_with_unsupported.num_tokens_in_messages(messages)
    print(result)



import pytest

from not_again_ai.base.file_system import readable_size


@pytest.mark.parametrize(
    ("size", "expected"),
    [
        (0, "0.00 B"),
        (523, "523.00 B"),
        (2048, "2.00 KB"),
        (5242880, "5.00 MB"),
        (10737418240, "10.00 GB"),
        (1099511627776, "1.00 TB"),
    ],
)
def test_human_readable_size(size: float, expected: str) -> None:
    assert readable_size(size) == expected, f"Failed for {size} bytes"



from typing import Any

import numpy as np
import numpy.typing as npt

from not_again_ai.statistics.dependence import pearson_correlation, pred_power_score_classification


def _example_1(rs: np.random.RandomState) -> tuple[npt.NDArray[Any], npt.NDArray[Any]]:
    """Example 1 - x is mostly predictive of y.
    x is categorical (strings), y is binary, both are numpy arrays
    """
    x0 = rs.choice(["a"], 200)
    y0 = rs.choice([0, 1], 200, p=[0.9, 0.1])

    x1 = rs.choice(["b"], 300)
    y1 = rs.choice([0, 1], 300, p=[0.1, 0.9])

    x = np.concatenate([x0, x1])
    y = np.concatenate([y0, y1])
    return (x, y)


def _example_2() -> tuple[npt.NDArray[Any], npt.NDArray[Any]]:
    """Example 2 - x completely predicts y"""
    x = np.array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0])
    y = np.array(["a", "a", "b", "b", "b", "b", "b", "b", "b", "b"])
    return (x, y)


def _example_3() -> tuple[npt.NDArray[Any], npt.NDArray[Any]]:
    """Example 3 - y contains all of the same values"""
    x = np.array([0, 0, 0, 0, 0, 0])
    y = np.array([3, 3, 3, 3, 3, 3])
    return (x, y)


def _example_4(rs: np.random.RandomState) -> tuple[npt.NDArray[Any], npt.NDArray[Any]]:
    """Example 4 - y is multi-class, x is numeric"""
    x0 = rs.normal(3, 0.1, 100)
    yo = rs.choice([0, 1, 2], 100, p=[0.9, 0.05, 0.05])

    x1 = rs.normal(2, 0.1, 100)
    y1 = rs.choice([0, 1, 2], 100, p=[0.05, 0.9, 0.05])

    x2 = rs.normal(1, 0.1, 100)
    y2 = rs.choice([0, 1, 2], 100, p=[0.05, 0.05, 0.9])

    x = np.concatenate([x0, x1, x2])
    y = np.concatenate([yo, y1, y2])
    return (x, y)


def _example_5(rs: np.random.RandomState) -> tuple[npt.NDArray[Any], npt.NDArray[Any]]:
    """Example 5 - x is not predictive of y (random noise)"""
    x = rs.choice(["a", "b"], 500)
    y = rs.choice([0, 1], 500)
    return (x, y)


def _example_6(rs: np.random.RandomState) -> tuple[npt.NDArray[Any], npt.NDArray[Any]]:
    """Example 6 - Both variables fully random: Correlation should be 0"""
    x = rs.randn(500)
    y = rs.randn(500)
    return (x, y)


def _example_7(rs: np.random.RandomState) -> tuple[npt.NDArray[Any], npt.NDArray[Any]]:
    """Example 7 - y = x^2 + noise"""
    x = (rs.rand(500) * 4) - 2
    y = x**2 + (rs.randn(500) * 0.2)
    return (x, y)


def _example_8(rs: np.random.RandomState) -> tuple[npt.NDArray[Any], npt.NDArray[Any]]:
    """Example 8 - y = x + noise, correlation should be  high"""
    x = rs.randn(500)
    e = rs.randn(500) * 0.2
    y = x + e
    return (x, y)


def test_pearson_correlation() -> None:
    rs = np.random.RandomState(365)

    x, y = _example_1(rs)
    res = pearson_correlation(x, y, is_x_categorical=True, is_y_categorical=True)
    assert res > 0.333

    x, y = _example_2()
    res = pearson_correlation(x, y, is_x_categorical=True, is_y_categorical=True)
    assert np.isclose(res, 1, atol=1e-6)

    x, y = _example_3()
    res = pearson_correlation(x, y, is_x_categorical=True, is_y_categorical=True, print_diagnostics=True)
    assert res == 1

    x, y = _example_4(rs)
    res = pearson_correlation(x, y, is_x_categorical=False, is_y_categorical=True)
    assert res > 0.333

    x, y = _example_5(rs)
    res = pearson_correlation(x, y, is_x_categorical=True, is_y_categorical=True)
    assert res < 0.333

    x, y = _example_6(rs)
    res = pearson_correlation(x, y)
    assert res < 0.333

    x, y = _example_7(rs)
    res = pearson_correlation(x, y)
    assert res >= 0

    x, y = _example_8(rs)
    res = pearson_correlation(x, y)
    assert res > 0.5


def test_pred_power_score_classification() -> None:
    rs = np.random.RandomState(365)

    x, y = _example_1(rs)
    res = pred_power_score_classification(x, y)
    assert res > 0.333

    x, y = _example_2()
    res = pred_power_score_classification(x, y)
    assert res == 1

    x, y = _example_3()
    res = pred_power_score_classification(x, y, print_diagnostics=True)
    assert res == 1

    x, y = _example_4(rs)
    res = pred_power_score_classification(x, y)
    assert res > 0.333

    x, y = _example_5(rs)
    res = pred_power_score_classification(x, y, cv_splits=10)
    assert res < 0.333



import importlib.util

if importlib.util.find_spec("playwright") is None:
    raise ImportError(
        "not_again_ai.data requires the 'data' extra to be installed. "
        "You can install it using 'pip install not_again_ai[data]'."
    )



from collections.abc import Callable
from multiprocessing.pool import ThreadPool
from typing import Any


def embarrassingly_parallel(
    func: Callable[..., Any],
    args_list: tuple[tuple[Any, ...], ...] | None,
    kwargs_list: list[dict[str, Any]] | None = None,
    num_processes: int = 1,
) -> list[Any]:
    """Call multiple functions in parallel providing either positional arguments, keyword arguments,
        or both. Return the function returns in a list ordered by order of the input arguments.

    If both are provided, positional and keyword arguments must be aligned in the same order
    and each list must be the same length.

    Args:
        func (Callable[..., Any]): Any function
        args_list (Optional[tuple[tuple[Any, ...], ...]]): A tuple of tuples each of positional arguments.
        kwargs_list (Optional[list[dict[str, Any]]], optional): A list of dictionaries containing keyword arguments. Defaults to None.
        num_processes (int, optional): Number of parallel processors to use. Defaults to 1.

    Raises:
        ValueError: If positional and keyword arguments are not aligned in the
            same order or if the lists are not the same length.
        ValueError: If neither positional nor keyword arguments are provided.

    Returns:
        list[Any]: list of the returns of each function call in order of the args_list or kwargs_list.
    """

    pool = ThreadPool(processes=num_processes)
    results = {}
    if (args_list is not None) and (kwargs_list is None):
        for idx, args in enumerate(args_list):
            results[idx] = pool.apply_async(func, args)
    elif (args_list is None) and (kwargs_list is not None):
        for idx, kwargs in enumerate(kwargs_list):
            results[idx] = pool.apply_async(func, kwds=kwargs)
    elif (args_list is not None) and (kwargs_list is not None):
        # in this case args_list and kwargs_list must be of the same length
        if len(args_list) == len(kwargs_list):
            for idx, (args, kwargs) in enumerate(zip(args_list, kwargs_list, strict=True)):
                results[idx] = pool.apply_async(func, args, kwargs)
        else:
            pool.close()
            pool.terminate()
            raise ValueError("args_list and kwargs_list must be of the same length")
    else:
        pool.close()
        pool.terminate()
        raise ValueError("either args_list or kwargs_list must be provided")

    return_results = []
    for _, res in results.items():
        return_results.append(res.get())

    pool.close()
    pool.terminate()
    return return_results


def embarrassingly_parallel_simple(funcs: list[Callable[..., Any]], num_processes: int = 1) -> list[Any]:
    """Executes the given functions in parallel and returns the results in the same order as the funcs were provided.

    Args:
        funcs (list[Callable[..., Any]]): A list of any functions that take no arguments.
        num_processes (int, optional): Number of parallel processors to use. Defaults to 1.

    Returns:
        list[Any]: list of the returns of each function call in order of the provided funcs.
    """

    pool = ThreadPool(processes=num_processes)
    results = {}
    for idx, func in enumerate(funcs):
        results[idx] = pool.apply_async(func)

    return_results = []
    for _, res in results.items():
        return_results.append(res.get())

    pool.close()
    pool.terminate()
    return return_results



import multiprocessing
import random
import time

import pytest

from not_again_ai.base.parallel import embarrassingly_parallel, embarrassingly_parallel_simple


def multby2(x: float, y: float, double: bool = False) -> float:
    time.sleep(random.uniform(0, 1))
    if double:
        return x * y * 2
    else:
        return x * y


def do_something() -> int:
    return 8


def do_something2() -> int:
    return 2


def echo(x: int) -> int:
    time.sleep(random.uniform(0, 1) / 10)
    return x


def test_embarrassingly_parallel() -> None:
    args = ((2, 2), (3, 3), (4, 4))

    result = embarrassingly_parallel(multby2, args, num_processes=multiprocessing.cpu_count())

    total = 0
    for x in result:
        total += x
    assert total == 4 + 9 + 16


def test_embarrassingly_parallel_both() -> None:
    args = ((2, 2), (3, 3), (4, 4))
    kwargs = [{"double": True}, {"double": False}, {"double": True}]

    result = embarrassingly_parallel(multby2, args, kwargs, num_processes=multiprocessing.cpu_count())

    total = 0
    for x in result:
        total += x
    assert total == 8 + 9 + 32


def test_embarrassingly_parallel_kwargs() -> None:
    kwargs = [{"x": 2, "y": 2, "double": True}, {"x": 3, "y": 3, "double": False}, {"x": 4, "y": 4, "double": True}]

    result = embarrassingly_parallel(multby2, None, kwargs, num_processes=multiprocessing.cpu_count())

    total = 0
    for x in result:
        total += x
    assert total == 8 + 9 + 32


def test_embarrassingly_parallel_exceptions() -> None:
    with pytest.raises(ValueError, match="either args_list or kwargs_list must be provided"):
        embarrassingly_parallel(multby2, None, None, num_processes=multiprocessing.cpu_count())

    args = ((2, 2), (3, 3), (4, 4))
    kwargs = [{"double": True}, {"double": False}]
    with pytest.raises(ValueError, match="args_list and kwargs_list must be of the same length"):
        embarrassingly_parallel(multby2, args, kwargs, num_processes=multiprocessing.cpu_count())


def test_embarrassingly_parallel_ordering() -> None:
    args = ((1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (10,))
    result = embarrassingly_parallel(echo, args, num_processes=3)
    assert result == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]


def test_embarrassingly_parallel_simple() -> None:
    result = embarrassingly_parallel_simple([do_something, do_something2], num_processes=2)
    assert result == [8, 2]
